import numpy as np
import pandas as pd
from ..sarracen_dataframe import SarracenDataFrame
from .utils import _get_mass, _get_origin
from .utils import _bin_particles_by_radius, _get_bin_midpoints

from typing import Tuple, Union


def azimuthal_average(data: 'SarracenDataFrame',
                      target: str,
                      r_in: Union[float, None] = None,
                      r_out: Union[float, None] = None,
                      bins: int = 300,
                      log: bool = False,
                      geometry: str = 'cylindrical',
                      origin: Union[list, None] = None,
                      retbins: bool = False) -> Union[np.ndarray,
                                                      Tuple[np.ndarray,
                                                            np.ndarray]]:
    """
    Calculates the 1D azimuthally-averaged profile for a target quantity.

    The profile is computed by segmenting the particles into radial bins
    (rings) and taking the mean of the target quantity from the particles
    within each bin.

    Parameters
    ----------
    data : SarracenDataFrame
        Particle data, in a SarracenDataFrame.
    target : str
        Column label of the target smoothing data.
    r_in : float, optional
        Inner radius of the disc. Defaults to the minimum r value.
    r_out : float, optional
        Outer radius of the disc. Defaults to the maximum r value.
    bins : int, optional
        Defines the number of equal-width bins in the range [r_in, r_out].
        Default is 300.
    log : bool, optional
        Whether to bin in log scale or not. Defaults to False.
    geometry : str, optional
        Coordinate system to use to calculate the particle radii. Can be
        either *spherical* or *cylindrical*. Defaults to *cylindrical*.
    origin : array-like, optional
        The x, y and z centre point around which to compute radii. Defaults to
        [0, 0, 0].
    retbins : bool, optional
        Whether to return the midpoints of the bins or not. Defaults to False.

    Returns
    -------
    array
        A NumPy array of length bins containing the averaged profile.
    array, optional
        The midpoint values of each bin. Only returned if *retbins=True*.

    Raises
    ------
    ValueError
        If the *geometry* is not *cylindrical* or *spherical*.
    """

    origin = _get_origin(origin)
    rbins, bin_edges = _bin_particles_by_radius(data, r_in, r_out, bins, log,
                                                geometry, origin)

    result = data[target].groupby(rbins).mean().to_numpy()

    if retbins:
        return result, _get_bin_midpoints(bin_edges, log)
    else:
        return result


def surface_density(data: 'SarracenDataFrame',
                    r_in: Union[float, None] = None,
                    r_out: Union[float, None] = None,
                    bins: int = 300,
                    log: bool = False,
                    geometry: str = 'cylindrical',
                    origin: Union[list, None] = None,
                    retbins: bool = False) -> Union[np.ndarray,
                                                    Tuple[np.ndarray,
                                                          np.ndarray]]:
    """
    Calculates the 1D azimuthally-averaged surface density profile.

    The surface density profile is computed by segmenting the particles into
    radial bins (rings) and dividing the total mass contained within each bin
    by the area of its respective ring.

    Parameters
    ----------
    data : SarracenDataFrame
        Particle data, in a SarracenDataFrame.
    r_in : float, optional
        Inner radius of the disc. Defaults to the minimum r value.
    r_out : float, optional
        Outer radius of the disc. Defaults to the maximum r value.
    bins : int, optional
        Defines the number of equal-width bins in the range [r_in, r_out].
        Default is 300.
    log : bool, optional
        Whether to bin in log scale or not. Defaults to False.
    geometry : str, optional
        Coordinate system to use to calculate the particle radii. Can be
        either *spherical* or *cylindrical*. Defaults to *cylindrical*.
    origin : array-like, optional
        The x, y and z centre point around which to compute radii. Defaults to
        [0, 0, 0].
    retbins : bool, optional
        Whether to return the midpoints of the bins or not. Defaults to False.

    Returns
    -------
    array
        A NumPy array of length bins containing the surface density profile.
    array, optional
        The midpoint values of each bin. Only returned if *retbins=True*.

    Raises
    ------
    ValueError
        If the *geometry* is not *cylindrical* or *spherical*.

    Notes
    -----
    The surface density averaging procedure for SPH is described in section
    3.2.6 of Lodato & Price (2010) [1]_.

    References
    ----------
    .. [1] G. Lodato & D. J. Price, "On the diffusive propagation of warps in
       thin accretion discs," MNRAS, 405, 2, 1212-1226 (2010).
       `doi:10.1111/j.1365-2966.2010.16526.x
       <https://doi.org/10.1111/j.1365-2966.2010.16526.x>`_

    """

    origin = _get_origin(origin)
    rbins, bin_edges = _bin_particles_by_radius(data, r_in, r_out, bins, log,
                                                geometry, origin)

    areas = np.pi * (bin_edges[1:] ** 2 - bin_edges[:-1] ** 2)

    mass = _get_mass(data)
    if isinstance(mass, pd.Series):
        sigma = mass.groupby(rbins).sum()
    else:
        sigma = data.groupby(rbins).count().iloc[:, 0] * mass

    if retbins:
        return (sigma / areas).to_numpy(), _get_bin_midpoints(bin_edges, log)
    else:
        return (sigma / areas).to_numpy()


def _calc_angular_momentum(data: 'SarracenDataFrame',
                           rbins: pd.Series,
                           origin: list,
                           unit_vector: bool) -> Tuple[pd.Series,
                                                       pd.Series,
                                                       pd.Series]:
    """
    Utility function to calculate angular momentum of the disc.

    Parameters
    ----------
    data: SarracenDataFrame
        Particle data, in a SarracenDataFrame.
    rbins: Series
        The radial bin to which each particle belongs.
    origin: list
        The x, y and z centre point around which to compute radii.
    unit_vector: bool
        Whether to convert the angular momentum to unit vectors.
        Default is True.

    Returns
    -------
    Lx, Ly, Lz: Series
        The x, y and z components of the angular momentum per bin.
    """

    mass = _get_mass(data)

    x_data = data[data.xcol].to_numpy() - origin[0]
    y_data = data[data.ycol].to_numpy() - origin[1]
    z_data = data[data.zcol].to_numpy() - origin[2]

    Lx = y_data * data[data.vzcol] - z_data * data[data.vycol]
    Ly = z_data * data[data.vxcol] - x_data * data[data.vzcol]
    Lz = x_data * data[data.vycol] - y_data * data[data.vxcol]

    if isinstance(mass, float):
        Lx = (mass * Lx).groupby(rbins).sum()
        Ly = (mass * Ly).groupby(rbins).sum()
        Lz = (mass * Lz).groupby(rbins).sum()
    else:
        Lx = (data[data.mcol] * Lx).groupby(rbins).sum()
        Ly = (data[data.mcol] * Ly).groupby(rbins).sum()
        Lz = (data[data.mcol] * Lz).groupby(rbins).sum()

    if unit_vector:
        Lmag = 1.0 / np.sqrt(Lx ** 2 + Ly ** 2 + Lz ** 2)

        Lx = Lx * Lmag
        Ly = Ly * Lmag
        Lz = Lz * Lmag

    return Lx, Ly, Lz


def angular_momentum(data: 'SarracenDataFrame',
                     r_in: Union[float, None] = None,
                     r_out: Union[float, None] = None,
                     bins: int = 300,
                     log: bool = False,
                     geometry: str = 'cylindrical',
                     origin: Union[list, None] = None,
                     retbins: bool = False,
                     unit_vector: bool = True) -> Union[Tuple[np.ndarray,
                                                              ...]]:
    """
    Calculates the angular momentum profile of the disc.

    The profile is computed by segmenting the particles into radial bins
    (rings) and summing the angular momentum of the particles within each
    bin.

    Parameters
    ----------
    data : SarracenDataFrame
        Particle data, in a SarracenDataFrame.
    r_in : float, optional
        Inner radius of the disc. Defaults to the minimum r value.
    r_out : float, optional
        Outer radius of the disc. Defaults to the maximum r value.
    bins : int, optional
        Defines the number of equal-width bins in the range [r_in, r_out].
        Default is 300.
    log : bool, optional
        Whether to bin in log scale or not. Defaults to False.
    geometry : str, optional
        Coordinate system to use to calculate the particle radii. Can be
        either *spherical* or *cylindrical*. Defaults to *cylindrical*.
    origin : array-like, optional
        The x, y and z centre point around which to compute radii. Defaults to
        [0, 0, 0].
    retbins : bool, optional
        Whether to return the midpoints of the bins or not. Defaults to False.
    unit_vector: bool, optional
        Whether to convert the angular momentum to unit vectors.
        Default is True.

    Returns
    -------
    array
        A NumPy array of length bins containing the angular momentum profile.
    array, optional
        The midpoint values of each bin. Only returned if *retbins=True*.

    Raises
    ------
    ValueError
        If the *geometry* is not *cylindrical* or *spherical*.
    """

    origin = _get_origin(origin)
    rbins, bin_edges = _bin_particles_by_radius(data, r_in, r_out, bins, log,
                                                geometry, origin)

    Lx_series, Ly_series, Lz_series = _calc_angular_momentum(data,
                                                             rbins,
                                                             origin,
                                                             unit_vector)
    Lx = Lx_series.to_numpy()
    Ly = Ly_series.to_numpy()
    Lz = Lz_series.to_numpy()

    if retbins:
        return Lx, Ly, Lz, _get_bin_midpoints(bin_edges, log)
    else:
        return Lx, Ly, Lz


def _calc_scale_height(data: 'SarracenDataFrame',
                       rbins: pd.Series,
                       origin: list) -> pd.Series:
    """
    Utility function to calculate the scale height of the disc.

    Parameters
    ----------
    data: SarracenDataFrame
        Particle data, in a SarracenDataFrame.
    rbins: Series
        The radial bin to which each particle belongs.
    origin : array-like, optional
        The x, y and z centre point around which to compute radii. Defaults to
        [0, 0, 0].

    Returns
    -------
    Series
        The scale height of the disc.
    """

    Lx, Ly, Lz = _calc_angular_momentum(data, rbins, origin, unit_vector=True)

    zdash = rbins.map(Lx).to_numpy() * data[data.xcol] \
        + rbins.map(Ly).to_numpy() * data[data.ycol] \
        + rbins.map(Lz).to_numpy() * data[data.zcol]

    return zdash.groupby(rbins).std()


def scale_height(data: 'SarracenDataFrame',
                 r_in: Union[float, None] = None,
                 r_out: Union[float, None] = None,
                 bins: int = 300,
                 log: bool = False,
                 geometry: str = 'cylindrical',
                 origin: Union[list, None] = None,
                 retbins: bool = False) -> Union[np.ndarray,
                                                 Tuple[np.ndarray,
                                                       np.ndarray]]:
    """
    Calculates the scale height, H/R, of the disc.

    The scale height, H/R, is computed by segmenting the particles into radial
    bins (rings) and calculating the angular momentum profile of the disc.
    Each particle takes the dot product of its position vector with the
    angular momentum vector of its corresponding bin. The standard deviation
    of this result per bin yields the scale height profile of the disc, which
    is divided by the midpoint radius of each bin.

    Parameters
    ----------
    data : SarracenDataFrame
        Particle data, in a SarracenDataFrame.
    r_in : float, optional
        Inner radius of the disc. Defaults to the minimum r value.
    r_out : float, optional
        Outer radius of the disc. Defaults to the maximum r value.
    bins : int, optional
        Defines the number of equal-width bins in the range [r_in, r_out].
        Default is 300.
    log : bool, optional
        Whether to bin in log scale or not. Defaults to False.
    geometry : str, optional
        Coordinate system to use to calculate the particle radii. Can be
        either *spherical* or *cylindrical*. Defaults to *cylindrical*.
    origin : array-like, optional
        The x, y and z centre point around which to compute radii. Defaults to
        [0, 0, 0].
    retbins : bool, optional
        Whether to return the midpoints of the bins or not. Defaults to False.

    Returns
    -------
    array
        A NumPy array of length bins scale height, H, profile.
    array, optional
        The midpoint values of each bin. Only returned if *retbins=True*.

    Raises
    ------
    ValueError
        If the *geometry* is not *cylindrical* or *spherical*.

    See Also
    --------
    :func:`angular_momentum` : Calculate the disc angular momentum profile.
    """

    origin = _get_origin(origin)
    rbins, bin_edges = _bin_particles_by_radius(data, r_in, r_out, bins, log,
                                                geometry, origin)

    midpoints = _get_bin_midpoints(bin_edges, log)
    H = _calc_scale_height(data, rbins, origin).to_numpy() / midpoints

    if retbins:
        return H, midpoints
    else:
        return H


def honH(data: 'SarracenDataFrame',
         r_in: Union[float, None] = None,
         r_out: Union[float, None] = None,
         bins: int = 300,
         log: bool = False,
         geometry: str = 'cylindrical',
         origin: Union[list, None] = None,
         retbins: bool = False) -> Union[np.ndarray, Tuple[np.ndarray,
                                                           np.ndarray]]:
    """
    Calculates <h>/H, the averaged smoothing length divided by the scale
    height.

    The profile is computed by segmenting the particles into radial bins
    (rings). The average smoothing length in each bin is divided by the scale
    height as calculated for that bin.

    Parameters
    ----------
    data : SarracenDataFrame
        Particle data, in a SarracenDataFrame.
    r_in : float, optional
        Inner radius of the disc. Defaults to the minimum r value.
    r_out : float, optional
        Outer radius of the disc. Defaults to the maximum r value.
    bins : int, optional
        Defines the number of equal-width bins in the range [r_in, r_out].
        Default is 300.
    log : bool, optional
        Whether to bin in log scale or not. Defaults to False.
    geometry : str, optional
        Coordinate system to use to calculate the particle radii. Can be
        either *spherical* or *cylindrical*. Defaults to *cylindrical*.
    origin : array-like, optional
        The x, y and z centre point around which to compute radii. Defaults to
        [0, 0, 0].
    retbins : bool, optional
        Whether to return the midpoints of the bins or not. Defaults to False.

    Returns
    -------
    array
        A NumPy array of length bins containing the <h>/H profile.
    array, optional
        The midpoint values of each bin. Only returned if *retbins=True*.

    Raises
    ------
    ValueError
        If the *geometry* is not *cylindrical* or *spherical*.

    See Also
    --------
    :func:`scale_height` : Calculate the scale height of a disc.
    """

    origin = _get_origin(origin)
    rbins, bin_edges = _bin_particles_by_radius(data, r_in, r_out, bins, log,
                                                geometry, origin)

    H = _calc_scale_height(data, rbins, origin).to_numpy()

    mean_h = data.groupby(rbins)[data.hcol].mean().to_numpy()

    if retbins:
        return mean_h / H, _get_bin_midpoints(bin_edges, log)
    else:
        return mean_h / H



from typing import Union, List

import pandas as pd

from ..sarracen_dataframe import SarracenDataFrame


def read_gradsph(filename: str,
                 separate_types: str = 'sinks') -> Union[List[
                                                         SarracenDataFrame],
                                                         SarracenDataFrame]:
    """
    Read data from a GradSPH dump file.

    Global values stored in the dump file are stored within the data frame in
    the dictionary ``params``.

    Parameters
    ----------
    filename : str
        Name of the file to be loaded.
    separate_types : {None, 'sinks', 'all'}, default='sinks'
        Whether to separate SPH particles and sink particles into separate
        SarracenDataFrames. ``None`` returns all particle types in one
        SarracenDataFrame. '`sinks`' and '`all`' separate sink particles into
        a second SarracenDataFrame.

    Returns
    -------
    SarracenDataFrame or list of SarracenDataFrame

    Examples
    --------
    By default, SPH particles are grouped into one SarracenDataFrame and sink
    particles into a second SarracenDataFrame.

    >>> sdf, sdf_sinks = sarracen.read_gradsph('col3139')
    """
    with open(filename, 'r') as fp:

        n_str, ninactive_str, nsink_str = fp.readline().split()
        n, ninactive, nsink = int(n_str), int(ninactive_str), int(nsink_str)

        t_str, gamma_str = fp.readline().split()
        t, gamma = float(t_str), float(gamma_str)

        params = {'n': n,
                  'ninactive': ninactive,
                  'nsink': nsink,
                  't': t,
                  'gamma': gamma}

        sinks = [fp.readline().split() for _ in range(nsink)]
        parts = [fp.readline().split() for _ in range(n - ninactive)]

        sink_header = ['x', 'y', 'z', 'vx', 'vy', 'vz', 'mass']
        part_header = ['x', 'y', 'z', 'vx', 'vy', 'vz', 'mass', 'h', 'cs',
                       'rho', 'temp']

        df_sinks = pd.DataFrame(sinks, columns=sink_header, dtype=float)
        df_parts = pd.DataFrame(parts, columns=part_header, dtype=float)

        if separate_types == 'sinks' or separate_types == 'all':
            df_list = [SarracenDataFrame(df_parts, params=params),
                       SarracenDataFrame(df_sinks, params=params)]
        else:
            df_list = [SarracenDataFrame(pd.concat([df_parts, df_sinks],
                                                   ignore_index=True),
                                         params=params)]

        return df_list[0] if len(df_list) == 1 else df_list



from typing import Any

import pandas as pd
from ..sarracen_dataframe import SarracenDataFrame


def read_csv(*args: Any, **kwargs: Any) -> SarracenDataFrame:
    """
    Read data from a csv file.

    Parameters
    ----------
    filename : str
        Name of the file to be loaded.

    Returns
    -------
    SarracenDataFrame
    """
    df = SarracenDataFrame(pd.read_csv(*args, **kwargs))

    df.units = _get_units(df.columns.to_series())
    df.columns = _get_labels(df.columns.to_series())

    return df


def _get_units(columns: pd.Series) -> pd.Series:
    return columns.str.extract(r'((?<=\[).+(?=\]))')[0]


def _get_labels(columns: pd.Series) -> pd.Series:
    return columns.str.extract(r'(^[^\[]*[^\s\[])')[0]



import numpy as np
import pandas as pd
import sys
from ..sarracen_dataframe import SarracenDataFrame
from typing import Tuple, Union


def _get_mass(data: 'SarracenDataFrame') -> Union[float, pd.Series]:
    if data.mcol is None:
        if 'mass' not in data.params:
            raise KeyError("'mass' column does not exist in this "
                           "SarracenDataFrame.")
        return data.params['mass']

    return data[data.mcol]


def _get_origin(origin: Union[list, None]) -> list:
    if origin is None:
        return [0.0, 0.0, 0.0]
    else:
        return origin


def _bin_particles_by_radius(data: 'SarracenDataFrame',
                             r_in: Union[float, None],
                             r_out: Union[float, None],
                             bins: int,
                             log: bool,
                             geometry: str,
                             origin: list) -> Tuple[pd.Series, np.ndarray]:
    """
    Utility function to bin particles in discrete intervals by radius.

    Parameters
    ----------
    data: SarracenDataFrame
        The particle dataset.
    r_in : float
        Inner radius of the disc. Defaults to the minimum r value.
    r_out : float
        Outer radius of the disc. Defaults to the maximum r value.
    bins : int
        Defines the number of equal-width bins in the range [r_in, r_out].
        Default is 300.
    log : bool
        Whether to bin in log scale or not. Defaults to False.
    geometry : str
        Coordinate system to use to calculate the particle radii. Can be
        either *spherical* or *cylindrical*. Defaults to *cylindrical*.
    origin : array-like
        The x, y and z centre point around which to compute radii. Defaults to
        [0, 0, 0].

    Returns
    -------
    rbins: Series
        The radial bin to which each particle belongs.
    bin_edges: ndarray
        Locations of the bin edges.
    """

    if geometry == 'spherical':
        r = np.sqrt((data[data.xcol] - origin[0]) ** 2
                    + (data[data.ycol] - origin[1]) ** 2
                    + (data[data.zcol] - origin[2]) ** 2)
    elif geometry == 'cylindrical':
        r = np.sqrt((data[data.xcol] - origin[0]) ** 2
                    + (data[data.ycol] - origin[1]) ** 2)
    else:
        raise ValueError("geometry should be either 'cylindrical' or "
                         "'spherical'")

    # should we add epsilon here?
    if r_in is None:
        r_in = r.min() - sys.float_info.epsilon
    if r_out is None:
        r_out = r.max() + sys.float_info.epsilon

    if log:
        bin_edges = np.logspace(np.log10(r_in), np.log10(r_out), bins+1)
    else:
        bin_edges = np.linspace(r_in, r_out, bins+1)
    rbins = pd.cut(r, pd.Series(bin_edges))

    return rbins, bin_edges


def _get_bin_midpoints(bin_edges: np.ndarray,
                       log: bool = False) -> np.ndarray:
    """
    Calculate the midpoint of bins given their edges.

    Parameters
    ----------
    bin_edges: ndarray
        Locations of the bin edges.
    log : bool, optional
        Whether to bin in log scale or not. Defaults to False.
    """

    if log:
        return np.sqrt(bin_edges[:-1] * bin_edges[1:])
    else:
        return 0.5 * (bin_edges[1:] - bin_edges[:-1]) + bin_edges[:-1]



from typing import Tuple, IO, Type

import numpy as np
import pandas as pd
from enum import IntEnum

from ..sarracen_dataframe import SarracenDataFrame


class MARISAIO_TAGS(IntEnum):
    startheader = 0
    endheader = 1
    startslice = 2
    endslice = 3
    luascript = 4
    n = 5
    ndim = 6
    dt = 7
    endt = 8
    t = 9
    gamma = 10
    G = 11
    epssq = 12
    munaught = 13
    viscbeta = 14
    alphaviscMax = 15
    alphaviscMin = 16
    alphamagMax = 17
    alphamagMin = 18
    alphauMax = 19
    alphauMin = 20
    integratorType = 21
    boundaryType = 22
    boundaryxl = 23
    boundaryxh = 24
    boundaryyl = 25
    boundaryyh = 26
    boundaryzl = 27
    boundaryzh = 28
    totalge = 29
    totalke = 30
    totalbe = 31
    totalue = 32
    totalpsie = 33
    totals = 34
    totalmomentum = 35
    rx = 36
    ry = 37
    rz = 38
    vx = 39
    vy = 40
    vz = 41
    bx = 42
    by = 43
    bz = 44
    m = 45
    P = 46
    rho = 47
    ue = 48
    ke = 49
    s = 50
    h = 51
    alpha = 52
    alphamag = 53
    alphau = 54
    fgx = 55
    fgy = 56
    fgz = 57
    fhx = 58
    fhy = 59
    fhz = 60
    fbx = 61
    fby = 62
    fbz = 63
    dbx = 64
    dby = 65
    dbz = 66
    ds = 67
    due = 68
    drhodh = 69
    dalpha = 70
    dalphamag = 71
    dalphau = 72
    curlv = 73
    curlvx = 74
    curlvy = 75
    curlvz = 76
    divv = 77
    curlb = 78
    curlbx = 79
    curlby = 80
    curlbz = 81
    divb = 82
    divbsymm = 83
    euleralpha = 84
    eulerbeta = 85
    vsigmax = 86
    psi = 87
    dpsi = 88
    id = 89
    nneigh = 90
    ax = 91
    ay = 92
    az = 93
    diva = 94
    psiheat = 95
    viscvsigmax = 96
    dataset = 97
    dustfrac = 98
    colour = 99

# 0: sets the reference point at the beginning of the file
# 1: sets the reference point at the current file position
# 2: sets the reference point at the end of the file


def _marisa_read_capture_pattern(fp: IO) -> Tuple[int, int]:
    if fp.read(7) != b"marisa\x00":
        raise AssertionError("Capture pattern not present. "
                             "Is this a valid data file?")
    version_major = int.from_bytes(fp.read(4), byteorder='little')
    version_minor = int.from_bytes(fp.read(4), byteorder='little')
    return version_major, version_minor


def _marisa_read_tag(fp: IO) -> int:
    tag = int.from_bytes(fp.read(4), byteorder='little')
    return tag


def _marisa_read_data(fp: IO, dtype: Type[np.number]) -> np.ndarray:
    size = int.from_bytes(fp.read(4), byteorder='little')
    data = fp.read(size)
    return np.frombuffer(data, dtype=dtype)


def _marisa_parse_tags(fp: IO) -> Tuple[np.ndarray, np.ndarray]:
    currentpos = fp.tell()

    fp.seek(0, 2)
    endpos = fp.tell()

    fp.seek(0, 0)

    _marisa_read_capture_pattern(fp)

    tags_list = []
    offsets_list = []

    while (fp.tell() < endpos):
        tag = _marisa_read_tag(fp)
        offset = fp.tell()
        size = int.from_bytes(fp.read(4), byteorder='little')
        fp.seek(size, 1)

        tags_list.append(tag)
        offsets_list.append(offset)

    tags = np.array(tags_list)
    offsets = np.array(offsets_list)

    fp.seek(currentpos, 0)

    return tags, offsets


def _marisa_count_slices(fp: IO, tags: np.ndarray) -> int:
    Ns = 0
    for tag in tags:
        if tag == MARISAIO_TAGS.startslice:
            Ns = Ns + 1
    return Ns


def read_marisa(filename: str,
                slicenumber: int = 0) -> SarracenDataFrame:
    """ Read data from a Marisa dump file.

    Parameters
    ----------
    filename : str
        Name of the file to be loaded.
    slicenumber : int, default=0
        The time slice to read from the data file.

    Returns
    -------
    SarracenDataFrame
    """
    fp = open(filename, "rb")
    tags, offsets = _marisa_parse_tags(fp)
    Ns = _marisa_count_slices(fp, tags)

    if slicenumber < 0:
        slicenumber = Ns + slicenumber

    if (slicenumber < 0 or slicenumber >= Ns):
        raise ValueError("Invalid slice number")

    verbose = False

    if verbose:
        print("seeking slice number: " + str(slicenumber))

    # find slice

    slicecounter = -1
    slicetagID = -1

    for i in range(len(tags)):

        if tags[i] == MARISAIO_TAGS.startslice:

            slicecounter = slicecounter + 1

            if (slicecounter == slicenumber):
                slicetagID = i

            if verbose:
                print("found slice: ")
                print("   slicecounter: " + str(slicecounter))
                print("   slicetagID: " + str(slicetagID))

        # all the header bit we will ignore

    if (slicetagID == -1 or slicecounter < slicenumber):
        raise ValueError("Slice number not found")

    done = False
    endsliceread = False
    i = slicetagID
    df = pd.DataFrame()
    params = dict()
    while not done:
        fp.seek(int(offsets[i]), 0)

        tag = tags[i]

        if (tag == MARISAIO_TAGS.endslice):
            endsliceread = True
            done = True
            i = len(tags) + 1
        elif (tag == MARISAIO_TAGS.n):
            params['n'] = _marisa_read_data(fp, np.int32)[0]
        elif (tag == MARISAIO_TAGS.t):
            params['t'] = _marisa_read_data(fp, np.float64)[0]
        elif (tag == MARISAIO_TAGS.totalge):
            params['totalge'] = _marisa_read_data(fp, np.float64)[0]
        elif (tag == MARISAIO_TAGS.totalke):
            params['totalke'] = _marisa_read_data(fp, np.float64)[0]
        elif (tag == MARISAIO_TAGS.totalue):
            params['totalue'] = _marisa_read_data(fp, np.float64)[0]
        elif (tag == MARISAIO_TAGS.totalbe):
            params['totalbe'] = _marisa_read_data(fp, np.float64)[0]
        elif (tag == MARISAIO_TAGS.totalpsie):
            params['totalpsie'] = _marisa_read_data(fp, np.float64)[0]
        elif (tag == MARISAIO_TAGS.totalmomentum):
            params['totalmomentum'] = _marisa_read_data(fp, np.float64)[0]

        if (tag == MARISAIO_TAGS.rx):
            df[MARISAIO_TAGS.rx.name] = _marisa_read_data(fp, np.double)
        if (tag == MARISAIO_TAGS.ry):
            df[MARISAIO_TAGS.ry.name] = _marisa_read_data(fp, np.double)
        if (tag == MARISAIO_TAGS.rz):
            df[MARISAIO_TAGS.rz.name] = _marisa_read_data(fp, np.double)
        if (tag == MARISAIO_TAGS.vx):
            df[MARISAIO_TAGS.vx.name] = _marisa_read_data(fp, np.double)
        if (tag == MARISAIO_TAGS.vy):
            df[MARISAIO_TAGS.vy.name] = _marisa_read_data(fp, np.double)
        if (tag == MARISAIO_TAGS.vz):
            df[MARISAIO_TAGS.vz.name] = _marisa_read_data(fp, np.double)

        if (tag == MARISAIO_TAGS.bx):
            df[MARISAIO_TAGS.bx.name] = _marisa_read_data(fp, np.double)
        if (tag == MARISAIO_TAGS.by):
            df[MARISAIO_TAGS.by.name] = _marisa_read_data(fp, np.double)
        if (tag == MARISAIO_TAGS.bz):
            df[MARISAIO_TAGS.bz.name] = _marisa_read_data(fp, np.double)
        if (tag == MARISAIO_TAGS.psi):
            df[MARISAIO_TAGS.psi.name] = _marisa_read_data(fp, np.double)

        if (tag == MARISAIO_TAGS.euleralpha):
            df[MARISAIO_TAGS.rx.name] = _marisa_read_data(fp, np.double)
        if (tag == MARISAIO_TAGS.ax):
            df[MARISAIO_TAGS.ax.name] = _marisa_read_data(fp, np.double)
        if (tag == MARISAIO_TAGS.ay):
            df[MARISAIO_TAGS.ay.name] = _marisa_read_data(fp, np.double)
        if (tag == MARISAIO_TAGS.az):
            df[MARISAIO_TAGS.az.name] = _marisa_read_data(fp, np.double)

        if (tag == MARISAIO_TAGS.m):
            df[MARISAIO_TAGS.m.name] = _marisa_read_data(fp, np.double)
        if (tag == MARISAIO_TAGS.h):
            df[MARISAIO_TAGS.h.name] = _marisa_read_data(fp, np.double)
        if (tag == MARISAIO_TAGS.rho):
            df[MARISAIO_TAGS.rho.name] = _marisa_read_data(fp, np.double)
        if (tag == MARISAIO_TAGS.P):
            df[MARISAIO_TAGS.P.name] = _marisa_read_data(fp, np.double)
        if (tag == MARISAIO_TAGS.ue):
            df[MARISAIO_TAGS.ue.name] = _marisa_read_data(fp, np.double)
        if (tag == MARISAIO_TAGS.ke):
            df[MARISAIO_TAGS.ke.name] = _marisa_read_data(fp, np.double)
        if (tag == MARISAIO_TAGS.s):
            df[MARISAIO_TAGS.s.name] = _marisa_read_data(fp, np.double)

        if (tag == MARISAIO_TAGS.alpha):
            df[MARISAIO_TAGS.alpha.name] = _marisa_read_data(fp, np.double)
        if (tag == MARISAIO_TAGS.alphamag):
            df[MARISAIO_TAGS.alphamag.name] = _marisa_read_data(fp, np.double)
        if (tag == MARISAIO_TAGS.alphau):
            df[MARISAIO_TAGS.alphau.name] = _marisa_read_data(fp, np.double)

        if (tag == MARISAIO_TAGS.divv):
            df[MARISAIO_TAGS.divv.name] = _marisa_read_data(fp, np.double)
        if (tag == MARISAIO_TAGS.divb):
            df[MARISAIO_TAGS.divb.name] = _marisa_read_data(fp, np.double)
        if (tag == MARISAIO_TAGS.divbsymm):
            df[MARISAIO_TAGS.divbsymm.name] = _marisa_read_data(fp, np.double)
        if (tag == MARISAIO_TAGS.curlb):
            df[MARISAIO_TAGS.curlb.name] = _marisa_read_data(fp, np.double)
        if (tag == MARISAIO_TAGS.dustfrac):
            df[MARISAIO_TAGS.dustfrac.name] = _marisa_read_data(fp, np.double)
        if (tag == MARISAIO_TAGS.colour):
            df[MARISAIO_TAGS.colour.name] = _marisa_read_data(fp, np.double)

        i = i + 1

    if (not endsliceread):
        raise AssertionError("Did not find end of slice tag")

    return SarracenDataFrame(df, params=params)



from typing import IO, Tuple, Type, Union, List, overload, Literal

import numpy as np
import pandas as pd

from ..sarracen_dataframe import SarracenDataFrame


def _read_fortran_block(fp: IO, bytesize: int) -> bytes:
    """ Helper function to read Fortran-written data.

    Fortran will add a 4-byte tag before and after any data writes. The value
    of this tag is equal to the number of bytes written. In our case, we do a
    simple sanity check that the start and end tag are consistent, but not
    validate the value of the tag with the size of the data read.
    """
    start_tag = fp.read(4)
    data = fp.read(bytesize)
    end_tag = fp.read(4)

    if (start_tag != end_tag):
        raise AssertionError("Fortran tags mismatch.")

    return data


def _read_capture_pattern(fp: IO) -> Tuple[Type[np.generic],
                                           Type[np.generic],
                                           int, bool]:
    """ Phantom dump validation plus default real and int sizes."""

    start_tag = fp.read(4)  # 4-byte Fortran tag

    def_types: List[Tuple[Type[np.generic],
                          Type[np.generic]]] = [(np.int32, np.float64),
                                                (np.int32, np.float32),
                                                (np.int64, np.float64),
                                                (np.int64, np.float32)]

    i1 = r1 = i2 = 0
    def_int_dtype, def_real_dtype = def_types[0]

    swap_endian = False
    for def_int_dtype, def_real_dtype in def_types:
        i1 = fp.read(def_int_dtype().itemsize)
        r1 = fp.read(def_real_dtype().itemsize)
        i2 = fp.read(def_int_dtype().itemsize)

        i1 = np.frombuffer(i1, count=1, dtype=def_int_dtype)[0]
        r1 = np.frombuffer(r1, count=1, dtype=def_real_dtype)[0]
        i2 = np.frombuffer(i2, count=1, dtype=def_int_dtype)[0]

        if (i1 == def_int_dtype(60769)
                and i2 == def_int_dtype(60878)
                and r1 == def_real_dtype(i2)):
            break
        if (i1.byteswap() == def_int_dtype(60769)
                and i2.byteswap() == def_int_dtype(60878)
                and r1.byteswap() == def_real_dtype(i2.byteswap())):
            swap_endian = True
            i1 = i1.byteswap()
            r1 = r1.byteswap()
            i2 = i2.byteswap()
            break
        else:  # rewind and try again
            fp.seek(-def_int_dtype().itemsize, 1)
            fp.seek(-def_real_dtype().itemsize, 1)
            fp.seek(-def_int_dtype().itemsize, 1)

    if (i1 != def_int_dtype(60769)
            or i2 != def_int_dtype(60878)
            or r1 != def_real_dtype(i2)):
        raise AssertionError("Could not determine default int or float "
                             "precision (i1, r1, i2 mismatch). "
                             "Is this a Phantom data file?")

    # iversion -- we don't actually check this
    iversion = fp.read(def_int_dtype().itemsize)
    iversion = np.frombuffer(iversion, count=1, dtype=def_int_dtype)[0]
    if swap_endian:
        iversion = iversion.byteswap()

    # integer 3 == 690706
    i3 = fp.read(def_int_dtype().itemsize)
    i3 = np.frombuffer(i3, count=1, dtype=def_int_dtype)[0]
    if swap_endian:
        i3 = i3.byteswap()
    if i3 != def_int_dtype(690706):
        raise AssertionError("Capture pattern error. i3 mismatch. "
                             "Is this a Phantom data file?")

    end_tag = fp.read(4)  # 4-byte Fortran tag

    # assert tags equal
    if (start_tag != end_tag):
        raise AssertionError("Capture pattern error. Fortran tags mismatch. "
                             "Is this a Phantom data file?")

    return def_int_dtype, def_real_dtype, iversion, swap_endian


def _read_file_identifier(fp: IO) -> str:
    """ Read the 100 character file identifier.

    The file identifier contains code version and date information.
    """
    return _read_fortran_block(fp, 100).decode('ascii').strip()


def _rename_duplicates(keys: list) -> list:
    seen = dict()

    for i, key in enumerate(keys):
        if key not in seen:
            seen[key] = 1
        else:
            seen[key] += 1
            keys[i] += f'_{seen[key]}'

    return keys


def _read_global_header_block(fp: IO,
                              dtype: Type[np.generic],
                              swap_endian: bool) -> Tuple[list, list]:
    nvars = np.frombuffer(_read_fortran_block(fp, 4), dtype=np.int32)[0]
    if swap_endian:
        nvars = nvars.byteswap()

    keys = []
    data = []

    if (nvars > 0):
        # each tag is 16 characters in length
        keys_str = _read_fortran_block(fp, 16*nvars).decode('ascii')
        keys = [keys_str[i:i+16].strip() for i in range(0, len(keys_str), 16)]

        raw_data = _read_fortran_block(fp, dtype().itemsize*nvars)
        data_np = np.frombuffer(raw_data, count=nvars, dtype=dtype)
        if swap_endian:
            data_np = data_np.byteswap()
        data = list(data_np)

    return keys, data


def _read_global_header(fp: IO,
                        def_int_dtype: Type[np.generic],
                        def_real_dtype: Type[np.generic],
                        swap_endian: bool) -> dict:
    """ Read global variables. """

    dtypes = [def_int_dtype, np.int8, np.int16, np.int32, np.int64,
              def_real_dtype, np.float32, np.float64]

    keys = []
    data = []
    for dtype in dtypes:
        new_keys, new_data = _read_global_header_block(fp, dtype, swap_endian)

        keys += new_keys
        data += new_data

    keys = _rename_duplicates(keys)

    global_vars = dict()
    for i in range(len(keys)):
        global_vars[keys[i]] = data[i]

    return global_vars


def _read_array_block(fp: IO,
                      df: pd.DataFrame,
                      n: int,
                      nums: np.ndarray,
                      def_int_dtype: Type[np.generic],
                      def_real_dtype: Type[np.generic],
                      swap_endian: bool) -> pd.DataFrame:

    dtypes = [def_int_dtype, np.int8, np.int16, np.int32, np.int64,
              def_real_dtype, np.float32, np.float64]

    for i in range(len(nums)):
        dtype = dtypes[i]
        for j in range(nums[i]):

            tag = _read_fortran_block(fp, 16).decode('ascii').strip()

            if tag in df.columns:
                count = 1
                original_tag = tag
                while tag in df.columns:
                    count += 1
                    tag = original_tag + f"_{count}"

            raw_data = _read_fortran_block(fp, dtype().itemsize * n)
            data: np.ndarray = np.frombuffer(raw_data, dtype=dtype)
            if swap_endian:
                data = data.byteswap()
            df[tag] = data

    return df


def _read_array_blocks(fp: IO,
                       def_int_dtype: Type[np.generic],
                       def_real_dtype: Type[np.generic],
                       mpi_blocks: int,
                       swap_endian: bool) -> Tuple[pd.DataFrame, pd.DataFrame]:
    """ Read particle data.

    Block 2 is always for sink particles. The number of MPI blocks is given by
    'nblocks' in the global header. The first quantity read in this function is
    the total number of blocks."""

    nblocks = np.frombuffer(_read_fortran_block(fp, 4), dtype=np.int32)[0]
    if swap_endian:
        nblocks = nblocks.byteswap()

    df = pd.DataFrame()
    df_sinks = pd.DataFrame()

    nblocks = int(nblocks / mpi_blocks)  # number of blocks per MPI process

    for j in range(0, mpi_blocks):

        n: List[int] = []
        nums: List[np.ndarray] = []

        for i in range(0, nblocks):
            start_tag = fp.read(4)

            n_val = np.frombuffer(fp.read(8), dtype=np.int64)[0]
            nums_val = np.frombuffer(fp.read(32), count=8, dtype=np.int32)
            if swap_endian:
                n_val = n_val.byteswap()
                nums_val = nums_val.byteswap()
            n.append(n_val)
            nums.append(nums_val)

            end_tag = fp.read(4)
            if (start_tag != end_tag):
                raise AssertionError("Fortran tags mismatch in array blocks.")

        df_tmp = pd.DataFrame()
        df_tmp_sinks = pd.DataFrame()

        for i in range(0, nblocks):
            # This assumes the second block is only for sink particles.
            # This is a valid assumption as this is what splash assumes.
            # For now we will just append sinks to the end of the data frame.

            # Can we avoid temporary df?
            if i == 1:
                # Not sure why, but it seems each MPI block repeats sinks
                df_tmp_sinks = _read_array_block(fp, df_tmp_sinks, n[i],
                                                 nums[i], def_int_dtype,
                                                 def_real_dtype, swap_endian)
            else:
                df_tmp = _read_array_block(fp, df_tmp, n[i], nums[i],
                                           def_int_dtype, def_real_dtype,
                                           swap_endian)

        df = pd.concat([df, df_tmp])
        df_sinks = pd.concat([df_sinks, df_tmp_sinks]).drop_duplicates()

    return df, df_sinks


def _create_mass_column(df: pd.DataFrame,
                        header_vars: dict) -> pd.DataFrame:
    """
    Creates a mass column with the mass of each particle when there are
    multiple itypes.
    """
    df['mass'] = header_vars['massoftype']
    for itype in df['itype'].unique():
        if itype > 1:
            mass = header_vars[f'massoftype_{itype}']
            df.loc[df.itype == itype, 'mass'] = mass
    return df


def _create_aprmass_column(df: pd.DataFrame,
                           header_vars: dict) -> pd.DataFrame:
    """
    Creates a mass column with the mass of each particle when there are
    multiple refinement levels.
    """
    df['mass'] = header_vars['massoftype']
    df['mass'] = df['mass']/(2**(df['apr_level'] - 1))

    return df


@overload
def read_phantom(filename: str,
                 separate_types: None,
                 ignore_inactive: bool = True) -> SarracenDataFrame: ...
@overload  # noqa: E302
def read_phantom(filename: str,
                 separate_types: Literal['sinks'] = 'sinks',
                 ignore_inactive: bool = True) -> Union[List[
                                                        SarracenDataFrame],
                                                        SarracenDataFrame]: ...
@overload  # noqa: E302
def read_phantom(filename: str,
                 separate_types: Literal['all'] = 'all',
                 ignore_inactive: bool = True) -> Union[List[
                                                        SarracenDataFrame],
                                                        SarracenDataFrame]: ...
def read_phantom(filename: str,  # noqa: E302
                 separate_types: Union[str, None] = 'sinks',
                 ignore_inactive: bool = True) -> Union[List[
                                                        SarracenDataFrame],
                                                        SarracenDataFrame]:
    """
    Read data from a Phantom dump file.

    This reads the native binary format of Phantom dump files, which in turn
    were derived from the binary file format used by sphNG.

    Global values stored in the dump file (time step, initial momentum, hfact,
    Courant factor, etc) are stored within the data frame in the dictionary
    ``params``.

    Parameters
    ----------
    filename : str
        Name of the file to be loaded.
    separate_types : {None, 'sinks', 'all'}, default='sinks'
        Whether to separate different particle types into several dataframes.
        ``None`` returns all particle types in one data frame. '`sinks`'
        separates sink particles into a second dataframe, and '`all`' returns
        all particle types in different dataframes.
    ignore_inactive : {True, False}, default=True
        If True, particles with negative smoothing length will not be read on
        import. These are typically particles that have been accreted onto a
        sink particle or are otherwise inactive.

    Returns
    -------
    SarracenDataFrame or list of SarracenDataFrame

    See Also
    --------
    :func:`SarracenDataFrame` : A pandas DataFrame with support for SPH data.

    Notes
    -----
    See the `Phantom documentation
    <https://phantomsph.readthedocs.io/en/latest/dumpfile.html>`_ for a full
    description of the Phantom binary file format.

    Examples
    --------
    By default, SPH particles are grouped into one data frame and sink
    particles into a second data frame.

    >>> import sarracen
    >>> sdf, sdf_sinks = sarracen.read_phantom('dumpfile_00000')

    A dump file containing multiple particle types, say gas + dust + sinks,
    can be separated into their own data frames by specifying
    ``separate_types='all'``.

    >>> sdf_g, sdf_d, sdf_sinks = sarracen.read_phantom('dumpfile_00000',
    ...                                                 separate_types='all')

    Global values are stored in the ``params`` dictionary.

    >>> sdf_g.params
    {'nparttot': np.int32(100000),
     'ntypes': np.int32(8),
     'npartoftype': np.int32(100000),
     'massoftype': np.float64(5.05e-7)
     'time': np.float64(0.05),
     'grainsize': np.float64(6.684491978609626e-14),
     'graindens': np.float64(5049628.378663718),
     'udist': np.float64(14960000000000.0)
     'umass': np.float64(1.9891e+33),
     'utime': np.float64(5022728.790082334),
     ...
    }

    """
    with open(filename, 'rb') as fp:
        def_int_dtype, def_real_dtype, iversion, swap_endian = \
            _read_capture_pattern(fp)
        file_identifier = _read_file_identifier(fp)

        header_vars = _read_global_header(fp, def_int_dtype, def_real_dtype,
                                          swap_endian)
        header_vars['file_identifier'] = file_identifier
        header_vars['iversion'] = iversion
        header_vars['def_int_dtype'] = def_int_dtype
        header_vars['def_real_dtype'] = def_real_dtype

        mpi_blocks = header_vars['nblocks'] if 'nblocks' in header_vars else 1

        df, df_sinks = _read_array_blocks(fp, def_int_dtype, def_real_dtype,
                                          mpi_blocks, swap_endian)

        if ignore_inactive and 'h' in df.columns:
            df = df[df['h'] > 0]

        # create mass column if multiple species in single dataframe
        if (separate_types != 'all'
                and 'itype' in df
                and df['itype'].nunique() > 1):
            df = _create_mass_column(df, header_vars)
        # create a column if APR is used and automatically scale masses
        elif 'apr_level' in df:
            df = _create_aprmass_column(df, header_vars)
        else:  # create global mass parameter
            header_vars['mass'] = header_vars['massoftype']

        df_list = []
        if separate_types == 'all':
            if 'itype' in df and df['itype'].nunique() > 1:
                for _, group in df.groupby('itype'):
                    itype = int(group["itype"].iloc[0])
                    mass_key = 'massoftype' if itype == 1 \
                        else f'massoftype_{itype}'
                    params = {**header_vars, **{"mass": header_vars[mass_key]}}
                    df_list.append(SarracenDataFrame(group.dropna(axis=1),
                                                     params=params))
            else:
                df_list = [SarracenDataFrame(df, params=header_vars)]

            if not df_sinks.empty:
                params = {key: value for key, value in header_vars.items()
                          if key != 'mass'}
                df_list.append(SarracenDataFrame(df_sinks,
                                                 params=params))

        elif separate_types == 'sinks':
            df_list = [SarracenDataFrame(df, params=header_vars)]
            if not df_sinks.empty:
                params = {key: value for key, value in header_vars.items()
                          if key != 'mass'}
                df_list.append(SarracenDataFrame(df_sinks,
                                                 params=params))
        else:
            df_list = [SarracenDataFrame(pd.concat([df, df_sinks],
                                                   ignore_index=True),
                                         params=header_vars)]

        return df_list[0] if len(df_list) == 1 else df_list



"""
Provides several rendering functions which produce matplotlib plots of SPH
data. These functions act as interfaces to interpolation functions within
interpolate.py.

These functions can be accessed directly, for example:
    render_2d(data, target)
Or, they can be accessed through a `SarracenDataFrame` object, for example:
    data.render_2d(target)
"""

from typing import Any, Union, Tuple

import numpy as np
import pandas as pd
from scipy.spatial.transform import Rotation
import seaborn as sns
from matplotlib import pyplot as plt
from matplotlib.axes import Axes
from matplotlib.colors import Colormap, LogNorm, SymLogNorm

from .interpolate import interpolate_2d_line, interpolate_2d, \
    interpolate_3d_proj, interpolate_3d_cross, interpolate_3d_vec, \
    interpolate_3d_cross_vec, interpolate_2d_vec, interpolate_3d_line
from .kernels import BaseKernel


def _default_axes(data: 'SarracenDataFrame',  # noqa: F821
                  x: Union[str, None],
                  y: Union[str, None]) -> Tuple[str, str]:
    """
    Utility function to determine the x & y columns to use for rendering.

    Parameters
    ----------
    data: SarracenDataFrame
        The particle dataset to render.
    x, y: str
        The x and y directional column labels passed to the render function.

    Returns
    -------
    x, y: str
        The directional column labels to use for rendering. Defaults to the
        x-column detected in `data`
    """
    if x is None:
        x = data.xcol
    if y is None:
        y = data.ycol

    return x, y


def _rotate_data(data: 'SarracenDataFrame',  # noqa: F821
                 x_data: np.ndarray,
                 y_data: np.ndarray,
                 z_data: np.ndarray,
                 rotation: Union[np.ndarray, list, Rotation, None],
                 rot_origin: Union[np.ndarray, list, pd.Series,
                                   str, None]) -> Tuple[np.ndarray,
                                                        np.ndarray,
                                                        np.ndarray]:
    """
    Rotate vector data in a particle dataset.

    Parameters
    ----------
    data: SarracenDataFrame
        The particle dataset to interpolate over.
    x_data, y_data, z_data: ndarray
        The directional vector data.
    rotation: array_like or SciPy Rotation
        The rotation to apply to the vector data. If defined as an array, the
        order of rotations is [z, y, x] in degrees
    rot_origin: array_like or ['com', 'midpoint']
        Point of rotation of the data. Only applies to 3D datasets. If
        array_like, then the [x, y, z] coordinates specify the point around
        which the data is rotated. If 'com', then data is rotated around the
        centre of mass. If 'midpoint', then data is rotated around the
        midpoint, that is, min + max / 2. Defaults to the midpoint.

    Returns
    -------
    x_data, y_data, z_data: ndarray
        The rotated x, y, and z directional data.
    """
    vectors = np.array([x_data, y_data, z_data]).transpose()
    if rotation is not None:
        if not isinstance(rotation, Rotation):
            rotation_obj = Rotation.from_euler('zyx',
                                               rotation,
                                               degrees=True)
        else:
            rotation_obj = rotation

        if rot_origin is None:
            # rot_origin = [0, 0, 0]
            rot_origin_arr = (vectors.min(0) + vectors.max(0)) / 2
        elif rot_origin == 'com':
            rot_origin_arr = data.centre_of_mass()
        elif rot_origin == 'midpoint':
            rot_origin_arr = (vectors.min(0) + vectors.max(0)) / 2
        elif not isinstance(rot_origin, (list, pd.Series, np.ndarray)):
            raise ValueError("rot_origin should be an [x, y, z] point or "
                             "'com' or 'midpoint'")
        elif len(rot_origin) != 3:
            raise ValueError("rot_origin should specify [x, y, z] point.")
        else:
            rot_origin_arr = rot_origin
        vectors = vectors - rot_origin_arr
        vectors = rotation_obj.apply(vectors)
        vectors = vectors + rot_origin_arr

    return vectors[:, 0], vectors[:, 1], vectors[:, 2]


def _default_bounding_box(data: 'SarracenDataFrame',  # noqa: F821
                          x: str,
                          y: str,
                          xlim: Union[Tuple[float, float], None],
                          ylim: Union[Tuple[float, float], None],
                          z_slice: Union[float, None] = None) -> np.ndarray:
    # boundaries of the plot default to the max & min values of the data.
    x_min = xlim[0] if xlim is not None and xlim[0] is not None else None
    y_min = ylim[0] if ylim is not None and ylim[0] is not None else None
    x_max = xlim[1] if xlim is not None and xlim[1] is not None else None
    y_max = ylim[1] if ylim is not None and ylim[1] is not None else None

    x_min = data.loc[:, x].min() if x_min is None else x_min
    y_min = data.loc[:, y].min() if y_min is None else y_min
    x_max = data.loc[:, x].max() if x_max is None else x_max
    y_max = data.loc[:, y].max() if y_max is None else y_max

    z_slice = 0 if z_slice is None else z_slice

    corners = [[x_min, y_min], [x_min, y_max], [x_max, y_min], [x_max, y_max]]
    for i in range(len(corners)):
        corners[i].append(z_slice)

    return np.array(corners).transpose()


def _default_bounds(data: 'SarracenDataFrame',  # noqa: F821
                    x_data: np.ndarray,
                    y_data: np.ndarray,
                    xlim: Union[Tuple[float, float], None],
                    ylim: Union[Tuple[float, float], None],
                    ) -> Tuple[Tuple[float, float],
                               Tuple[float, float]]:
    """
    Utility function to determine the 2-dimensional boundaries to use in 2D
    rendering.

    Parameters
    ----------
    data: SarracenDataFrame
        The particle dataset to render.
    x, y: str
        The directional column labels that will be used for rendering.
    xlim, ylim: tuple of float
        The minimum and maximum values passed to the render function, in
        particle data space.

    Returns
    -------
    xlim, ylim: tuple of float
        The minimum and maximum values to use for rendering, in particle data
        space. Defaults to the maximum and minimum values of `x` and `y`,
        snapped to the nearest integer.
    """
    # boundaries of the plot default to the max & min values of the data.
    x_min = xlim[0] if xlim is not None and xlim[0] is not None else None
    y_min = ylim[0] if ylim is not None and ylim[0] is not None else None
    x_max = xlim[1] if xlim is not None and xlim[1] is not None else None
    y_max = ylim[1] if ylim is not None and ylim[1] is not None else None

    x_min = min(x_data) if x_min is None else x_min
    y_min = min(y_data) if y_min is None else y_min
    x_max = max(x_data) if x_max is None else x_max
    y_max = max(y_data) if y_max is None else y_max

    return (x_min, x_max), (y_min, y_max)


def _set_pixels(x_pixels: Union[int, None],
                y_pixels: Union[int, None],
                xlim: Tuple[float, float],
                ylim: Tuple[float, float],
                default: int) -> Tuple[int, int]:
    """
    Utility function to determine the number of pixels to interpolate over in
    2D interpolation.

    Parameters
    ----------
    x_pixels, y_pixels: int
        The number of pixels in the x & y directions passed to the
        interpolation function.
    xlim, ylim: tuple of float
        The minimum and maximum values to use in interpolation, in particle
        data space.

    Returns
    -------
    x_pixels, y_pixels: int
        The number of pixels in the x & y directions to use in 2D
        interpolation.
    """
    # set # of pixels to maintain an aspect ratio that is the same as the
    # underlying bounds of the data.

    dx = xlim[1] - xlim[0]
    dy = ylim[1] - ylim[0]

    if y_pixels is None:
        if x_pixels is None:
            x_pixels = default
        y_pixels = int(np.rint(x_pixels * (dy / dx)))
    elif x_pixels is None:
        x_pixels = int(np.rint(y_pixels * (dx / dy)))

    return x_pixels, y_pixels


def render(data: 'SarracenDataFrame',  # noqa: F821
           target: str,
           x: Union[str, None] = None,
           y: Union[str, None] = None,
           z: Union[str, None] = None,
           xsec: Union[float, None] = None,
           kernel: Union[BaseKernel, None] = None,
           x_pixels: Union[int, None] = None,
           y_pixels: Union[int, None] = None,
           xlim: Union[Tuple[float, float], None] = None,
           ylim: Union[Tuple[float, float], None] = None,
           cmap: Union[str, Colormap] = 'gist_heat',
           cbar: bool = True,
           cbar_kws: dict = {},
           cbar_ax: Union[Axes, None] = None,
           ax: Union[Axes, None] = None,
           exact: bool = False,
           backend: Union[str, None] = None,
           integral_samples: int = 1000,
           rotation: Union[np.ndarray, list, Rotation, None] = None,
           rot_origin: Union[np.ndarray, list, str, None] = None,
           log_scale: bool = False,
           symlog_scale: bool = False,
           dens_weight: Union[bool, None] = None,
           normalize: bool = True,
           hmin: bool = False,
           corotation: Union[np.ndarray, list, None] = None,
           **kwargs: Any) -> Axes:
    """
    Render a scalar SPH target variable to a grid plot.

    Parameters
    ----------
    data : SarracenDataFrame
        Particle data, in a SarracenDataFrame.
    target: str
        Column label of the target variable.
    x, y, z: str, optional
        Column labels of the x, y & z directional axes. Defaults to the columns
        detected in `data`.
    xsec: float, optional.
        For a 3D dataset, the z value to take a cross-section at. If none,
        column interpolation is performed.
    kernel: BaseKernel, optional
        Kernel to use for smoothing the target data. Defaults to the kernel
        specified in `data`.
    x_pixels, y_pixels: int, optional
        Number of pixels present in the final image.
    xlim, ylim: tuple of float, optional
        The starting and ending corners of the final 2D image.
    vmin, vmax: float, optional
        Lower and upper limits of the range of values for the colour bar.
    cmap: str or Colormap, optional
        The color map to use when plotting a 2D image.
    cbar: bool, optional
        True if a colorbar should be drawn.
    cbar_kws: dict, optional
        Keyword arguments to pass to matplotlib.figure.Figure.colorbar().
    cbar_ax: Axes, optional
        Axes to draw the colorbar in, if not provided then space will be taken
        from the main Axes.
    ax: Axes, optional
        The main axes in which to draw the rendered image.
    exact: bool, optional
        Whether to use exact interpolation of the data. For cross-sections this
        is ignored. Defaults to False.
    backend: ['cpu', 'gpu'], optional
        The computation backend to use when interpolating this data. Defaults
        to 'gpu' if CUDA is enabled, otherwise 'cpu' is used. A manually
        specified backend in `data` will override the default.
    integral_samples: int, optional
        If using column interpolation, the number of sample points to take when
        approximating the 2D column kernel.
    rotation: array_like or SciPy Rotation, optional
        The rotation to apply to the data before interpolation. If defined as
        an array, the order of rotations is [z, y, x] in degrees. Only applies
        to 3D datasets.
    rot_origin: array_like or ['com', 'midpoint'], optional
        Point of rotation of the data. Only applies to 3D datasets. If
        array_like, then the [x, y, z] coordinates specify the point around
        which the data is rotated. If 'com', then data is rotated around the
        centre of mass. If 'midpoint', then data is rotated around the
        midpoint, that is, min + max / 2. Defaults to the midpoint.
    log_scale: bool, optional
        Whether to use a logarithmic scale for color coding.
    symlog_scale: bool, optional
        Whether to use a symmetrical logarithmic scale for color coding (i.e.,
        allows positive and negative values). Optionally add "linthresh" and
        "linscale" to kwargs to set the linear region and the scaling of linear
        values, respectively (defaults to 1e-9 and 1, respectively). Only works
        if log_scale == True.
    dens_weight: bool, optional
        If True, will plot the target multiplied by the density. Defaults to
        True for column-integrated views and False for everything else.
    normalize: bool, optional
        If True, will normalize the interpolation. Defaults to True.
    hmin: bool, optional
        If True, a minimum smoothing length of 0.5 * pixel size will be
        imposed. This ensures each particle contributes to at least one grid
        cell / pixel. Defaults to False (this may change in a future verison).
    corotation: list, optional
        Moves particles to the co-rotating frame of two location. corotation
        contains two lists which correspond to the two x, y, z coordinates.
    kwargs: other keyword arguments
        Keyword arguments to pass to ax.imshow.

    Returns
    -------
    Axes
        The resulting matplotlib axes, which contain the 2d rendered image.

    Raises
    ------
    ValueError
        If `x_pixels` or `y_pixels` are less than or equal to zero, or if the
        specified `x` and `y` minimum and maximums result in an invalid region.
    KeyError
        If `target`, `x`, `y`, mass, density, or smoothing length columns do
        not exist in `data`.

    Notes
    -----
    The standard render will interpolate the target quantity, :math:`A`, from
    the particles to a pixel grid using the following equation:

        .. math::

            A_{pixel} = \\sum_b \\frac{m_b}{\\rho_b} A_b W_{ab}(h_b)

    where :math:`m` is the mass, :math:`\\rho` is the density, and :math:`W` is
    the smoothing kernel with smoothing length, :math:`h`.

    Normalized interpolation divides the above summation by an interpolation of
    a constant scalar field equal to 1:

        .. math::

            A_{pixel} = \\frac{\\sum_b \\frac{m_b}{\\rho_b} A_b W_{ab}(h_b)}
                              {\\sum_b \\frac{m_b}{\\rho_b} W_{ab}(h_b)}

    In theory, the denominator will be equal to 1 and dividing by 1 has no
    impact. In practice, the particle arrangement and the smoothing kernel
    affects the quality of interpolation. Normalizing by this approximation of
    1 helps to account for this.

    For when to use normalized interpolation, the advice given by Splash is
    recommended: in general use it for smoother renderings, but avoid when
    there are free surfaces, as it can cause them to be over-exaggerated.

    Density-weighted interpolation will interpolate the quantity
    :math:`\\rho A`, that is, the target :math:`A` multiplied by the density,
    :math:`\\rho`. If normalize=True, then density-weighted interpolation will
    be normalized by the density.

    Column-integrated views of 3D data (i.e., xsec=None) will calculate the
    following:

        .. math::

            A_{pixel} = \\sum_b \\frac{m_b}{\\rho_b} A_b \\int W_{ab}(h_b) dz ,

    which uses the integral of the kernel along the chosen line of sight.

    Exact rendering calculates the volume integral of the kernel through each
    pixel using the method of Petkova et al (2018) [1]_. It only works for the
    cubic spline kernel.

    References
    ----------
    .. [1] M. A. Petkova, G. Laibe & I. A. Bonnell, "Fast and accurate Voronoi
       density gridding from Lagrangian hydrodynamics data," J. Comput. Phys.,
       353, 15, 300-315 (2018). `doi:10.1016/j.jcp.2017.10.024
       <https://doi.org/10.1016/j.jcp.2017.10.024>`_

    """
    if data.get_dim() == 2:
        if dens_weight is None:
            dens_weight = False
        img = interpolate_2d(data, target, x, y, kernel, x_pixels, y_pixels,
                             xlim, ylim, exact, backend, dens_weight,
                             normalize, hmin)
    elif data.get_dim() == 3:
        if xsec is not None:
            if dens_weight is None:
                dens_weight = False
            img = interpolate_3d_cross(data, target, x, y, z, xsec, kernel,
                                       corotation, rotation, rot_origin,
                                       x_pixels, y_pixels, xlim, ylim, backend,
                                       dens_weight, normalize, hmin)
        else:
            img = interpolate_3d_proj(data, target, x, y, kernel,
                                      integral_samples, corotation, rotation,
                                      rot_origin, x_pixels, y_pixels, xlim,
                                      ylim, exact, backend, dens_weight,
                                      normalize, hmin)
    else:
        raise ValueError('`data` is not a valid number of dimensions.')

    if ax is None:
        ax = plt.gca()

    x, y = _default_axes(data, x, y)
    corners = _default_bounding_box(data, x, y, xlim, ylim, xsec)
    rotated_corners = _rotate_data(data, corners[0], corners[1], corners[2],
                                   rotation, rot_origin)
    xlim, ylim = _default_bounds(data, rotated_corners[0], rotated_corners[1],
                                 xlim, ylim)

    kwargs.setdefault("origin", 'lower')
    kwargs.setdefault("extent", [xlim[0], xlim[1], ylim[0], ylim[1]])
    if log_scale:
        # By default, a log scale plot will only cover 4 levels of magnitude.
        vminref = 10 ** (np.log10(kwargs.get("vmax", img.max())) - 4)
        vmin = kwargs.get('vmin', max(vminref, img.min()))

        if symlog_scale:
            kwargs.setdefault("norm",
                              SymLogNorm(kwargs.pop("linthresh", 1e-9),
                                         linscale=kwargs.pop("linscale", 1.),
                                         vmin=vmin,
                                         vmax=kwargs.get('vmax')))
        else:
            kwargs.setdefault("norm", LogNorm(clip=True,
                                              vmin=vmin,
                                              vmax=kwargs.get('vmax')))
        kwargs.pop("vmin", None)
        kwargs.pop("vmax", None)

    graphic = ax.imshow(img, cmap=cmap, **kwargs)
    if rotation is not None and data.get_dim() == 3:
        if corotation is not None:
            ax.set_xlabel(x)
            ax.set_ylabel(y)
    else:
        ax.set_xlabel(x)
        ax.set_ylabel(y)

    if cbar:
        colorbar = ax.figure.colorbar(graphic, cbar_ax, ax, **cbar_kws)
        if 'label' not in cbar_kws:
            label = target
            if data.get_dim() == 3 and xsec is None:
                label = f"column {label}"
            if log_scale:
                label = f"log ({label})"
            colorbar.ax.set_ylabel(label)

    return ax


def lineplot(data: 'SarracenDataFrame',  # noqa: F821
             target: str,
             x: Union[str, None] = None,
             y: Union[str, None] = None,
             z: Union[str, None] = None,
             kernel: Union[BaseKernel, None] = None,
             pixels: int = 512,
             xlim: Union[Tuple[float, float], None] = None,
             ylim: Union[Tuple[float, float], None] = None,
             zlim: Union[Tuple[float, float], None] = None,
             ax: Union[Axes, None] = None,
             backend: Union[str, None] = None,
             log_scale: bool = False,
             dens_weight: bool = False,
             normalize: bool = True,
             hmin: bool = False,
             **kwargs: Any) -> Axes:
    """
    Render a scalar SPH target variable to line plot.

    Parameters
    ----------
    data : SarracenDataFrame
        Particle data, in a SarracenDataFrame.
    target: str
        Column label of the target variable.
    x, y, z: str, optional
        Column labels of the x, y & z directional axes. Defaults to the columns
        detected in `data`.
    kernel: BaseKernel, optional
        Kernel to use for smoothing the target data. Defaults to the kernel
        specified in `data`.
    pixels: int, optional
        Number of samples taken across the x axis in the final plot.
    xlim, ylim, zlim: tuple of float, optional
        Coordinates of the two points that make up the cross-sectional line.
    ax: Axes, optional
        The main axes in which to draw the final plot.
    backend: ['cpu', 'gpu'], optional
        The computation backend to use when interpolating this data. Defaults
        to 'gpu' if CUDA is enabled, otherwise 'cpu' is used. A manually
        specified backend in `data` will override the default.
    log_scale: bool, optional
        Whether to use a logarithmic scale for color coding.
    dens_weight: bool, optional
        If True, will plot the target mutliplied by the density. Defaults to
        False.
    normalize: bool, optional
        If True, will normalize the interpolation. Defaults to False (this may
        change in future versions).
    hmin: bool, optional
        If True, a minimum smoothing length of 0.5 * pixel size will be
        imposed. This ensures each particle contributes to at least one grid
        cell / pixel. Defaults to False (this may change in a future verison).
    kwargs: other keyword arguments
        Keyword arguments to pass to sns.lineplot.

    Returns
    -------
    Axes
        The resulting matplotlib axes, which contain the 2d rendered image.

    Raises
    -------
    ValueError
        If `x_pixels` or `y_pixels` are less than or equal to zero, or if the
        specified `x` and `y` minimum and maximums result in an invalid region.
    KeyError
        If `target`, `x`, `y`, mass, density, or smoothing length columns do
        not exist in `data`.
    """

    if data.get_dim() == 2:
        img = interpolate_2d_line(data, target, x, y, kernel, pixels, xlim,
                                  ylim, backend, dens_weight, normalize, hmin)
    else:
        img = interpolate_3d_line(data, target, x, y, z, kernel, pixels, xlim,
                                  ylim, zlim, backend, dens_weight, normalize,
                                  hmin)

    if isinstance(xlim, float) or isinstance(xlim, int):
        xlim = xlim, xlim
    if isinstance(ylim, float) or isinstance(ylim, int):
        ylim = ylim, ylim

    x, y = _default_axes(data, x, y)
    xlim, ylim = _default_bounds(data, data.loc[:, x], data.loc[:, y],
                                 xlim, ylim)

    if data.get_dim() == 2:
        upper_lim = np.sqrt((xlim[1] - xlim[0])**2 + (ylim[1] - ylim[0])**2)
    else:
        if z is None:
            z = data.zcol
        if z not in data.columns:
            raise KeyError(f"z-directional column '{z}' does not exist in the "
                           f"provided dataset.")

        if isinstance(zlim, float) or isinstance(zlim, int):
            zlim = zlim, zlim
        zmin = data.loc[:, z].min()
        z1 = zmin if zlim is None or zlim[0] is None else zlim[0]
        z2 = zmin if zlim is None or zlim[1] is None else zlim[1]
        zlim = z2, z1

        upper_lim = np.sqrt((xlim[1] - xlim[0])**2
                            + (ylim[1] - ylim[0])**2
                            + (zlim[1] - zlim[0])**2)

    ax = sns.lineplot(x=np.linspace(0, upper_lim, img.size),
                      y=img,
                      ax=ax, **kwargs)

    if log_scale:
        ax.set(yscale='log')

    ax.margins(x=0, y=0)

    label = f'({x}, {y})' if data.get_dim() == 2 else f'({x}, {y}, {z})'
    ax.set_xlabel('cross-section ' + label)

    label = target
    if log_scale:
        label = f"log ({label})"
    ax.set_ylabel(label)

    return ax


def streamlines(data: 'SarracenDataFrame',  # noqa: F821
                target: Union[Tuple[str, str], Tuple[str, str, str]],
                x: Union[str, None] = None,
                y: Union[str, None] = None,
                z: Union[str, None] = None,
                xsec: Union[float, None] = None,
                kernel: Union[BaseKernel, None] = None,
                integral_samples: int = 1000,
                rotation: Union[np.ndarray, list, Rotation, None] = None,
                rot_origin: Union[np.ndarray, list, str, None] = None,
                x_pixels: Union[int, None] = None,
                y_pixels: Union[int, None] = None,
                xlim: Union[Tuple[float, float], None] = None,
                ylim: Union[Tuple[float, float], None] = None,
                ax: Union[Axes, None] = None,
                exact: bool = False,
                backend: Union[str, None] = None,
                dens_weight: Union[bool, None] = None,
                normalize: bool = True,
                hmin: bool = False,
                **kwargs: Any) -> Axes:
    """
    Create an SPH interpolated streamline plot of a target vector.

    Render the data within a SarracenDataFrame to a 2D matplotlib object, by
    rendering the values of a target vector. The contributions of all particles
    near the rendered area are summed and stored to a 2D grid for the x & y
    axes of the target vector. This data is then used to create a streamline
    plot using ax.streamlines().

    Parameters
    ----------
    data: SarracenDataFrame
        Particle data, in a SarracenDataFrame.
    target: str tuple of shape (2) or (3).
        Column label of the target vector. Shape must match the # of dimensions
        in `data`.
    x, y, z: str, optional
        Column label of the x, y & z directional axes. Defaults to the columns
        detected in `data`.
    xsec: float, optional
        The z to take a cross-section at. If none, column interpolation is
        performed.
    kernel: BaseKernel, optional
        Kernel to use for smoothing the target data. Defaults to the kernel
        specified in `data`.
    integral_samples: int, optional
        If using column interpolation, the number of sample points to take when
        approximating the 2D column kernel.
    rotation: array_like or SciPy Rotation, optional
        The rotation to apply to the data before interpolation. If defined as
        an array, the order of rotations is [z, y, x] in degrees.
    rot_origin: array_like or ['com', 'midpoint'], optional
        Point of rotation of the data. Only applies to 3D datasets. If
        array_like, then the [x, y, z] coordinates specify the point around
        which the data is rotated. If 'com', then data is rotated around the
        centre of mass. If 'midpoint', then data is rotated around the
        midpoint, that is, min + max / 2. Defaults to the midpoint.
    x_pixels, y_pixels: int, optional
        Number of interpolation samples to pass to ax.streamlines(). Default
        values are chosen to keep a consistent aspect ratio.
    xlim, ylim: float, optional
        The minimum and maximum values to use in interpolation, in particle
        data space. Defaults to the minimum and maximum values of `x` and `y`.
    ax: Axes, optional
        The main axes in which to draw the rendered image.
    exact: bool, optional
        Whether to use exact interpolation of the data. For cross-sections
        this is ignored. Defaults to False.
    backend: ['cpu', 'gpu'], optional
        The computation backend to use when interpolating this data. Defaults
        to 'gpu' if CUDA is enabled, otherwise 'cpu' is used. A manually
        specified backend in `data` will override the default.
    dens_weight: bool, optional
        If True, will plot the target mutliplied by the density. Defaults to
        True for column-integrated views and False for everything else.
    normalize: bool, optional
        If True, will normalize the interpolation. Defaults to False (this may
        change in future versions).
    hmin: bool, optional
        If True, a minimum smoothing length of 0.5 * pixel size will be
        imposed. This ensures each particle contributes to at least one grid
        cell / pixel. Defaults to False (this may change in a future verison).
    kwargs: other keyword arguments
        Keyword arguments to pass to ax.streamlines()

    Returns
    -------
    Axes
        The resulting matplotlib axes, which contains the streamline plot.

    Raises
    ------
    ValueError
        If `x_pixels` or `y_pixels` are less than or equal to zero, or if the
        specified `x` and `y` minimum and maximums result in an invalid region,
        or if the number of dimensions in the target vector does not match the
        data, or if `data` is not 2 or 3 dimensional.
    KeyError
        If `target`, `x`, `y`, `z` (for 3-dimensional data), mass, density, or
        smoothing length columns do not exist in `data`.
    """
    # Choose between the various interpolation functions available, based on
    # initial data passed to this function.
    if data.get_dim() == 2:
        if not len(target) == 2:
            raise ValueError('Target vector is not 2-dimensional.')
        if dens_weight is None:
            dens_weight = False
        img = interpolate_2d_vec(data, target[0], target[1], x, y, kernel,
                                 x_pixels, y_pixels, xlim, ylim, exact,
                                 backend, dens_weight, normalize, hmin)
    elif data.get_dim() == 3:
        if not len(target) == 3:
            raise ValueError('Target vector is not 3-dimensional.')
        if xsec is not None:
            if dens_weight is None:
                dens_weight = False
            img = interpolate_3d_cross_vec(data, target[0], target[1],
                                           target[2], xsec, x, y, z, kernel,
                                           rotation, rot_origin, x_pixels,
                                           y_pixels, xlim, ylim, backend,
                                           dens_weight, normalize, hmin)
        else:
            if dens_weight is None:
                dens_weight = True
            img = interpolate_3d_vec(data, target[0], target[1], target[2], x,
                                     y, kernel, integral_samples, rotation,
                                     rot_origin, x_pixels, y_pixels, xlim,
                                     ylim, exact, backend, dens_weight,
                                     normalize, hmin)
    else:
        raise ValueError('`data` is not a valid number of dimensions.')

    if ax is None:
        ax = plt.gca()

    x, y = _default_axes(data, x, y)
    corners = _default_bounding_box(data, x, y, xlim, ylim, xsec)
    rotated_corners = _rotate_data(data, corners[0], corners[1], corners[2],
                                   rotation, rot_origin)
    xlim, ylim = _default_bounds(data, rotated_corners[0], rotated_corners[1],
                                 xlim, ylim)

    kwargs.setdefault("color", 'black')
    ax.streamplot(np.linspace(xlim[0], xlim[1], np.size(img[0], 1)),
                  np.linspace(ylim[0], ylim[1], np.size(img[0], 0)),
                  img[0], img[1], **kwargs)

    ax.set_xlim(xlim)
    ax.set_ylim(ylim)

    # remove the x & y labels if the data is rotated, since these no longer
    # have physical relevance to the displayed data.
    if rotation is None:
        ax.set_xlabel(x)
        ax.set_ylabel(y)

    return ax


def arrowplot(data: 'SarracenDataFrame',  # noqa: F821
              target: Union[Tuple[str, str], Tuple[str, str, str]],
              x: Union[str, None] = None,
              y: Union[str, None] = None,
              z: Union[str, None] = None,
              xsec: Union[float, None] = None,
              kernel: Union[BaseKernel, None] = None,
              integral_samples: int = 1000,
              rotation: Union[np.ndarray, list, Rotation, None] = None,
              rot_origin: Union[np.ndarray, list, str, None] = None,
              x_arrows: Union[int, None] = None,
              y_arrows: Union[int, None] = None,
              xlim: Union[Tuple[float, float], None] = None,
              ylim: Union[Tuple[float, float], None] = None,
              ax: Union[Axes, None] = None,
              qkey: bool = True,
              qkey_kws: Union[dict, None] = None,
              exact: bool = False,
              backend: Union[str, None] = None,
              dens_weight: Union[bool, None] = None,
              normalize: bool = True,
              hmin: bool = False,
              **kwargs: Any) -> Axes:
    """
    Create an SPH interpolated vector field plot of a target vector.

    Render the data within a SarracenDataFrame to a 2D matplotlib object, by
    rendering the values of a target vector. The contributions of all particles
    near the rendered area are summed and stored to a 2D grid for the x & y
    axes of the target vector. This data is then used to create an arrow plot
    using ax.quiver().

    Parameters
    ----------
    data : SarracenDataFrame
        Particle data, in a SarracenDataFrame.
    target: str tuple of shape (2) or (3).
        Column label of the target vector. Shape must match the # of dimensions
        in `data`.
    x, y, z: str, optional
        Column label of the x, y & z directional axes. Defaults to the columns
        detected in `data`.
    xsec: float, optional
        The z to take a cross-section at. If none, column interpolation is
        performed.
    kernel: BaseKernel, optional
        Kernel to use for smoothing the target data. Defaults to the kernel
        specified in `data`.
    integral_samples: int, optional
        If using column interpolation, the number of sample points to take when
        approximating the 2D column kernel.
    rotation: array_like or SciPy Rotation, optional
        The rotation to apply to the data before interpolation. If defined as
        an array, the order of rotations is [z, y, x] in degrees.
    rot_origin: array_like or ['com', 'midpoint'], optional
        Point of rotation of the data. Only applies to 3D datasets. If
        array_like, then the [x, y, z] coordinates specify the point around
        which the data is rotated. If 'com', then data is rotated around the
        centre of mass. If 'midpoint', then data is rotated around the
        midpoint, that is, min + max / 2. Defaults to the midpoint.
    x_arrows, y_arrows: int, optional
        Number of arrows in the output image in the x & y directions. Default
        values are chosen to keep a consistent aspect ratio.
    xlim, ylim: tuple of float, optional
        The minimum and maximum values to use in interpolation, in particle
        data space. Defaults to the minimum and maximum values of `x` and `y`.
    ax: Axes, optional
        The main axes in which to draw the rendered image.
    qkey: bool, optional
        Whether to include a quiver key on the final plot.
    qkey_kws: dict, optional
        Keywords to pass through to ax.quiver.
    exact: bool, optional
        Whether to use exact interpolation of the data. For cross-sections this
        is ignored. Defaults to False.
    backend: ['cpu', 'gpu'], optional
        The computation backend to use when interpolating this data. Defaults
        to 'gpu' if CUDA is enabled, otherwise 'cpu' is used. A manually
        specified backend in `data` will override the default.
    dens_weight: bool, optional
        If True, will plot the target mutliplied by the density. Defaults to
        True for column-integrated views and False for everything else.
    normalize: bool, optional
        If True, will normalize the interpolation. Defaults to False (this may
        change in future versions).
    hmin: bool, optional
        If True, a minimum smoothing length of 0.5 * pixel size will be
        imposed. This ensures each particle contributes to at least one grid
        cell / pixel. Defaults to False (this may change in a future verison).
    kwargs: other keyword arguments
        Keyword arguments to pass to ax.quiver()

    Returns
    -------
    Axes
        The resulting matplotlib axes, which contains the arrow plot.

    Raises
    ------
    ValueError
        If `x_arrows` or `y_arrows` are less than or equal to zero, or if the
        specified `x` and `y` minimum and maximums result in an invalid region,
        or if the number of dimensions in the target vector does not match the
        data, or if `data` is not 2 or 3-dimensional.
    KeyError
        If `target`, `x`, `y`, `z` (for 3-dimensional data), mass, density, or
        smoothing length columns do not exist in `data`.
    """
    x, y = _default_axes(data, x, y)
    corners = _default_bounding_box(data, x, y, xlim, ylim, xsec)
    rotated_corners = _rotate_data(data, corners[0], corners[1], corners[2],
                                   rotation, rot_origin)
    xlim, ylim = _default_bounds(data, rotated_corners[0], rotated_corners[1],
                                 xlim, ylim)
    x_arrows, y_arrows = _set_pixels(x_arrows, y_arrows, xlim, ylim, 20)

    if data.get_dim() == 2:
        if not len(target) == 2:
            raise ValueError('Target vector is not 2-dimensional.')
        if dens_weight is None:
            dens_weight = False
        img = interpolate_2d_vec(data, target[0], target[1], x, y, kernel,
                                 x_arrows, y_arrows, xlim, ylim, exact,
                                 backend, dens_weight, normalize, hmin)
    elif data.get_dim() == 3:
        if not len(target) == 3:
            raise ValueError('Target vector is not 3-dimensional.')
        if xsec is not None:
            if dens_weight is None:
                dens_weight = False
            img = interpolate_3d_cross_vec(data, target[0], target[1],
                                           target[2], xsec, x, y, z, kernel,
                                           rotation, rot_origin, x_arrows,
                                           y_arrows, xlim, ylim, backend,
                                           dens_weight, normalize, hmin)
        else:
            if dens_weight is None:
                dens_weight = True
            img = interpolate_3d_vec(data, target[0], target[1], target[2], x,
                                     y, kernel, integral_samples, rotation,
                                     rot_origin, x_arrows, y_arrows, xlim,
                                     ylim, exact, backend, dens_weight,
                                     normalize, hmin)
    else:
        raise ValueError('`data` is not a valid number of dimensions.')

    if ax is None:
        ax = plt.gca()

    kwargs.setdefault("angles", 'uv')
    kwargs.setdefault("pivot", 'mid')
    ax.quiver(np.linspace(xlim[0], xlim[1], np.size(img[0], 1)),
              np.linspace(ylim[0], ylim[1], np.size(img[0], 0)),
              img[0], img[1], **kwargs)
    Q = ax.quiver(np.linspace(xlim[0], xlim[1], np.size(img[0], 1)),
                  np.linspace(ylim[0], ylim[1], np.size(img[0], 0)),
                  img[0], img[1], **kwargs)

    if qkey:
        if qkey_kws is None:
            qkey_kws = dict()
        # approximately equivalent to the top right of the plot.
        qkey_kws.setdefault('X', 0.85)
        qkey_kws.setdefault('Y', 1.02)

        # find a reasonable default value for the quiver key length.
        key_length = np.mean(np.sqrt(img[0] ** 2 + img[1] ** 2))
        key_length = float(np.format_float_positional(key_length,
                                                      precision=1,
                                                      unique=False,
                                                      fractional=False,
                                                      trim='k'))
        qkey_kws.setdefault('U', key_length)
        qkey_kws.setdefault('label', f"= {qkey_kws['U']}")

        qkey_kws.setdefault('labelpos', 'E')
        qkey_kws.setdefault('coordinates', 'axes')

        ax.quiverkey(Q, **qkey_kws)

    # remove the x & y labels if the data is rotated, since these no longer
    # have physical relevance to the displayed data.
    if rotation is None:
        ax.set_xlabel(x)
        ax.set_ylabel(y)

    ax.set_xlim(xlim)
    ax.set_ylim(ylim)

    return ax
